@@ -26,6 +26,10 @@
 #include <glib-object.h>
 #include <gnokii.h>
 
+#include <bluetooth/bluetooth.h>
+#include <bluetooth/sdp.h>
+#include <bluetooth/sdp_lib.h>
+
 #include "phonemgr-utils.h"
 
 /* This is the hash table containing information about driver <-> device
@@ -95,14 +99,98 @@ phonemgr_utils_gn_error_to_string (gn_error error, PhoneMgrError *perr)
 	}
 }
 
-static gboolean
+gboolean
 phonemgr_utils_is_bluetooth (const char *addr)
 {
 	return (g_file_test (addr, G_FILE_TEST_EXISTS) == FALSE);
 }
 
+static int
+get_rfcomm_channel (sdp_record_t *rec)
+{
+	int channel = -1;
+	sdp_list_t *protos = NULL;
+
+	if (sdp_get_access_protos (rec, &protos) != 0)
+		goto end;
+
+	channel = sdp_get_proto_port (protos, RFCOMM_UUID);
+ end:
+	sdp_list_foreach(protos, (sdp_list_func_t)sdp_list_free, 0);
+	sdp_list_free(protos, 0);
+	return channel;
+}
+
+/* Determine whether the given device supports Dial-Up Networking, and if so
+ * what the RFCOMM channel number for the service is.
+ */
+static int
+find_dun_channel (bdaddr_t *adapter, bdaddr_t *device)
+{
+	sdp_session_t *sdp = NULL;
+	sdp_list_t *search = NULL, *attrs = NULL, *recs = NULL, *tmp;
+	uuid_t browse_uuid, dun_id, obex_uuid;
+	uint16_t proto_desc_list = SDP_ATTR_PROTO_DESC_LIST;
+	uint16_t svclass_id_list = SDP_ATTR_SVCLASS_ID_LIST;
+	int channel = -1;
+
+	sdp = sdp_connect (adapter, device, SDP_RETRY_IF_BUSY);
+	if (!sdp)
+		goto end;
+
+	/* Normally, we'd just do a search for OBEX_FILETRANS_SVCLASS_ID,
+	 * but we also want to check multiple services.  So instead, we
+	 * browse all rfcomm obex services.
+	 */
+	sdp_uuid16_create(&browse_uuid, PUBLIC_BROWSE_GROUP);
+	sdp_uuid16_create(&dun_id, DIALUP_NET_SVCLASS_ID);
+	search = sdp_list_append (NULL, &browse_uuid);
+	search = sdp_list_append (search, &dun_id);
+
+	attrs = sdp_list_append (NULL, &proto_desc_list);
+	attrs = sdp_list_append (attrs, &svclass_id_list);
+
+	if (sdp_service_search_attr_req (sdp, search,
+					 SDP_ATTR_REQ_INDIVIDUAL, attrs,
+					 &recs))
+		goto end;
+
+	for (tmp = recs; tmp != NULL; tmp = tmp->next) {
+		sdp_record_t *rec = tmp->data;
+
+		/* If this service is better than what we've
+		 * previously seen, try and get the channel number.
+		 */
+		channel = get_rfcomm_channel (rec);
+		if (channel > 0)
+			goto end;
+	}
+
+end:
+	sdp_list_free (recs, (sdp_free_func_t)sdp_record_free);
+	sdp_list_free (search, NULL);
+	sdp_list_free (attrs, NULL);
+	sdp_close(sdp);
+
+	return channel;
+}
+
+int
+phonemgr_utils_get_channel (char *device)
+{
+	bdaddr_t src, dst;
+	int channel;
+
+	bacpy (&src, BDADDR_ANY);
+	str2ba(device, &dst);
+
+	channel = find_dun_channel (&src, &dst);
+
+	return channel;
+}
+
 char *
-phonemgr_utils_write_config (const char *driver, const char *addr)
+phonemgr_utils_write_config (const char *driver, const char *addr, int channel)
 {
 	if (phonemgr_utils_is_bluetooth (addr) == FALSE) {
 		return g_strdup_printf ("[global]\n"
@@ -110,10 +198,21 @@ phonemgr_utils_write_config (const char *driver, const char *addr)
 			"model = %s\n"
 			"connection = serial\n", addr, driver);
 	} else {
-		return g_strdup_printf ("[global]\n"
-				"port = %s\n"
-				"model = %s\n"
-				"connection = bluetooth\n", addr, driver);
+		if (channel > 0) {
+			return g_strdup_printf ("[global]\n"
+						"port = %s\n"
+						"model = %s\n"
+						"connection = bluetooth\n"
+						"channel = %d\n",
+						addr,
+						driver,
+						channel);
+		} else {
+			return g_strdup_printf ("[global]\n"
+						"port = %s\n"
+						"model = %s\n"
+						"connection = bluetooth\n", addr, driver);
+		}
 	}
 }
 
@@ -156,11 +255,11 @@ phonemgr_utils_driver_for_device (const char *device)
 
 static void
 phonemgr_utils_driver_parse_start_tag (GMarkupParseContext *ctx,
-		const gchar         *element_name,
-		const gchar        **attr_names,
-		const gchar        **attr_values,
-		gpointer             data,
-		GError             **error)
+				       const char          *element_name,
+				       const char         **attr_names,
+				       const char         **attr_values,
+				       gpointer             data,
+				       GError             **error)
 {
 	const char *phone_name, *phone_driver;
 
@@ -253,16 +352,14 @@ phonemgr_utils_init_hash_tables (void)
 	phonemgr_utils_start_parse ();
 }
 
-#define MODEL_SIZE 64
 char *
 phonemgr_utils_guess_driver (PhonemgrState *phone_state, const char *device,
 			     GError **error)
 {
-	char model[MODEL_SIZE];
+	const char *model;
 	char *driver;
 	gn_error err;
 
-	memset (model, 0, MODEL_SIZE);
 	phonemgr_utils_init_hash_tables ();
 
 	driver = phonemgr_utils_driver_for_device (device);
@@ -272,16 +369,11 @@ phonemgr_utils_guess_driver (PhonemgrState *phone_state, const char *device,
 	if (phone_state == NULL)
 		return NULL;
 
-	gn_data_clear(&phone_state->data);
-	phone_state->data.model = model;
+	model = gn_lib_get_phone_model (&phone_state->state);
 
-	err = gn_sm_functions(GN_OP_Identify,
-			&phone_state->data, &phone_state->state);
-
-	if (err != GN_ERR_NONE) {
+	if (model == NULL) {
 		PhoneMgrError perr;
-		g_warning ("gn_sm_functions: %s",
-				phonemgr_utils_gn_error_to_string (err, &perr));
+		g_warning ("gn_lib_get_phone_model failed");
 		goto bail;
 	}
 
@@ -295,22 +387,29 @@ bail:
 }
 
 PhonemgrState *
-phonemgr_utils_connect (const char *device, const char *driver, GError **error)
+phonemgr_utils_connect (const char *device, const char *driver, int channel, GError **error)
 {
-	PhonemgrState *phonemgr_state = NULL;
+	PhonemgrState *phone_state = NULL;
 	char *config, **lines;
 	gn_data data;
 	struct gn_statemachine state;
 	gn_error err;
 
-	config = phonemgr_utils_write_config (driver ? driver : PHONEMGR_DEFAULT_DRIVER, device);
+	if (phonemgr_utils_is_bluetooth (device) != FALSE) {
+		if (phonemgr_utils_connection_is_supported (PHONEMGR_CONNECTION_BLUETOOTH) == FALSE) {
+			//FIXME set the error message
+			return NULL;
+		}
+	}
+
+	config = phonemgr_utils_write_config (driver ? driver : PHONEMGR_DEFAULT_DRIVER, device, channel);
 	lines = g_strsplit (config, "\n", -1);
 	g_free (config);
 
 	if (gn_cfg_memory_read ((const char **)lines) < 0) {
 		g_warning ("gn_cfg_memory_read");
 		g_strfreev (lines);
-		goto bail;
+		return NULL;
 	}
 	g_strfreev (lines);
 
@@ -318,40 +417,117 @@ phonemgr_utils_connect (const char *device, const char *driver, GError **error)
 
 	if (gn_cfg_phone_load("", &state) < 0) {
 		g_warning ("gn_cfg_phone_load");
-		goto bail;
+		return NULL;
 	}
 
 	err = gn_gsm_initialise(&state);
 	if (err != GN_ERR_NONE) {
 		PhoneMgrError perr;
 		g_warning ("gn_gsm_initialise: %s",
-				phonemgr_utils_gn_error_to_string (err, &perr));
-		goto bail;
+			   phonemgr_utils_gn_error_to_string (err, &perr));
+		return NULL;
 	}
 
-	phonemgr_state = g_new (PhonemgrState, 1);
-	phonemgr_state->data = data;
-	phonemgr_state->state = state;
+	phone_state = g_new (PhonemgrState, 1);
+	phone_state->data = data;
+	phone_state->state = state;
 
-bail:
-
-	return phonemgr_state;
+	return phone_state;
 }
 
 void
-phonemgr_utils_disconnect (PhonemgrState *state)
+phonemgr_utils_disconnect (PhonemgrState *phone_state)
 {
-	g_return_if_fail (state != NULL);
+	g_return_if_fail (phone_state != NULL);
 
-	gn_sm_functions (GN_OP_Terminate, NULL, &state->state);
-	phonemgr_utils_gn_statemachine_clear (&state->state);
-	gn_data_clear (&state->data);
+	gn_lib_phone_close (&phone_state->state);
+	phonemgr_utils_gn_statemachine_clear (&phone_state->state);
+	gn_data_clear (&phone_state->data);
 }
 
 void
-phonemgr_utils_free (PhonemgrState *state)
+phonemgr_utils_free (PhonemgrState *phone_state)
 {
-	g_return_if_fail (state != NULL);
-	g_free (state);
+	g_return_if_fail (phone_state != NULL);
+	g_free (phone_state);
+}
+
+void
+phonemgr_utils_tell_driver (const char *addr)
+{
+	GError *error = NULL;
+	PhonemgrState *phone_state;
+	const char *model;
+	int channel;
+
+	channel = phonemgr_utils_get_channel (addr);
+	phone_state = phonemgr_utils_connect (addr, NULL, channel, &error);
+	if (phone_state == NULL) {
+		g_warning ("Couldn't connect to the '%s' phone: %s", addr, PHONEMGR_CONDERR_STR(error));
+		if (error != NULL)
+			g_error_free (error);
+		return;
+	}
+
+	model = gn_lib_get_phone_model (&phone_state->state);
+
+	g_print ("model: '%s'\n", model);
+
+	phonemgr_utils_disconnect (phone_state);
+	phonemgr_utils_free (phone_state);
+}
+
+time_t
+gn_timestamp_to_gtime (gn_timestamp stamp)
+{
+	GDate *date;
+	time_t time = 0;
+	int i;
+
+	if (gn_timestamp_isvalid (stamp) == FALSE)
+		return time;
+
+	date = g_date_new_dmy (stamp.day, stamp.month, stamp.year);
+	for (i = 1970; i < stamp.year; i++) {
+		if (g_date_is_leap_year (i) != FALSE)
+			time += 3600 * 24 * 366;
+		else
+			time += 3600 * 24 * 365;
+	}
+	time += g_date_get_day_of_year (date) * 3600 * 24;
+	time += stamp.hour * 3600 + stamp.minute * 60 + stamp.second;
+
+	g_free (date);
+
+	return time;
+}
+
+gboolean
+phonemgr_utils_connection_is_supported (PhonemgrConnectionType type)
+{
+	gn_connection_type conntype;
+
+	switch (type) {
+	case PHONEMGR_CONNECTION_BLUETOOTH:
+		conntype = GN_CT_Bluetooth;
+		break;
+	case PHONEMGR_CONNECTION_SERIAL:
+		conntype = GN_CT_Serial;
+		break;
+	case PHONEMGR_CONNECTION_IRDA:
+		conntype = GN_CT_Irda;
+		if (!gn_lib_is_connectiontype_supported (conntype))
+			conntype = GN_CT_Infrared;
+		break;
+	case PHONEMGR_CONNECTION_USB:
+		conntype = GN_CT_DKU2LIBUSB;
+		if (!gn_lib_is_connectiontype_supported (conntype))
+			conntype = GN_CT_DKU2;
+		break;
+	default:
+		g_assert_not_reached ();
+	}
+
+	return gn_lib_is_connectiontype_supported (conntype);
 }
 
